// Utility Functions used for the application

#include "stdafx.h"

#include "parasolid_kernel.h"
#include "kernel_interface.h"
#include "hc.h"
#include "hp_bridge.h"

#include <math.h>
#include <assert.h>

#include "htools.h"
#include "HBaseOperator.h"


// we must use a customer normalize function since Parasolid requires double precision
// and HOOPS Compute_Normalized_Vector supports single precision
static void normalize_vector(double* vector)
{
	double length;
	length = sqrt(vector[0]*vector[0]+vector[1]*vector[1]+vector[2]*vector[2]);
	vector[0]/=length;
	vector[1]/=length;
	vector[2]/=length;
}

static void Compute_Double_Precision_Cross_Product(double * vector1, double * vector2, double * up)
{
	up[0] = vector1[1] * vector2[2]  -  vector1[2] * vector2[1];
    up[1] = vector1[2] * vector2[0]  -  vector1[0] * vector2[2];
    up[2] = vector1[0] * vector2[1]  -  vector1[1] * vector2[0];
	normalize_vector(up);
}


// compute the contrustion-plane contrained vectors used for geometry insertion
void Get_Viewplane_Vectors(double * view, double * up)
{
	Point	position, target;
	Vector	up_vector, view_vector, orthogonal;
	float	x, y, z;
	double	ortho[3];

	HC_Show_Net_Camera_Up_Vector(&x, &y, &z);
	up_vector.Set(x, y, z);

	HC_Show_Net_Camera_Position (&x, &y, &z);
	position.Set(x, y, z);

	HC_Show_Net_Camera_Target (&x, &y, &z);
	target.Set(x, y, z);

	view_vector.Set(position.x - target.x,
					position.y - target.y,
					position.z - target.z);

	// find vector in plane orthogonal to view and up plane
	HC_Compute_Cross_Product(&view_vector, &up_vector, &orthogonal);

	// now do the casting to double
	view[0] = (double) position.x - (double) target.x;
	view[1] = (double) position.y - (double) target.y;
	view[2] = (double) position.z - (double) target.z;
	normalize_vector(view);

	ortho[0] = (double) orthogonal.x;
	ortho[1] = (double) orthogonal.y;
	ortho[2] = (double) orthogonal.z;
	normalize_vector(ortho);

	Compute_Double_Precision_Cross_Product(view, ortho, up);
}


// insert a circle into the contruction plane using the viewpoint
// to world xform to determine the points
void Create_Axis_Circle (Point & center,float radius,int axis)
{
	Point points[3], temp;
	
	float angle = 2.0f * PI / 3.0f;

	// create three points on the circle in the appropriate view plane

	if (axis == XZ) {
		for (int i = 0; i < 3; i++) {
			temp.Set(center.x + radius * (float)cos (i * angle),
							center.y,
							center.z + radius * (float)sin (i * angle));
			HC_Compute_Coordinates(".", "viewpoint", &temp, "world", &temp);
			points[i] = temp;
			}
		HC_Insert_Circle(&points[0], &points[1], &points[2]);
	}

	else if (axis == XY) {
		for (int i = 0; i < 3; i++) {
			temp.Set(center.x + radius * (float)cos (i * angle),
							center.y + radius * (float)sin (i * angle),
							center.z);
			HC_Compute_Coordinates(".", "viewpoint", &temp, "world", &temp);
			points[i] = temp; 
			}
		HC_Insert_Circle(&points[0], &points[1], &points[2]);
	}

	else if (axis == YZ) {
		for (int i = 0; i < 3; i++) {
			temp.Set(center.x,
							center.y + radius * (float)cos (i * angle),
							center.z + radius * (float)sin (i * angle));
			HC_Compute_Coordinates(".", "viewpoint", &temp, "world", &temp);
			points[i] = temp;
			}
		HC_Insert_Circle(&points[0], &points[1], &points[2]);
	}
}


// represent a wireframe sphere using three circles
void Create_Wireframe_Sphere (Point & center, float radius)
{
	Create_Axis_Circle(center, radius, XY);
	Create_Axis_Circle(center, radius, XZ);
	Create_Axis_Circle(center, radius, YZ);
}


// insert a wireframe cylinder into the contruction plane using the viewpoint
// to world xform to determine the points
void Create_Wireframe_Cylinder(Point center, float radius, float height)
{
	Point	bottom, top;
	
	bottom = center;
	Create_Axis_Circle(bottom, radius, XZ);

	top.Set(bottom.x, bottom.y + height, bottom.z);
	Create_Axis_Circle(top, radius, XZ);

	float angle = 2.0f * PI / 8.0f;

	for (int i = 0; i < 8; i++) {
		Point top, bottom;
		top.Set(center.x + radius * (float)cos (i * angle),
						center.y,
						center.z + radius * (float)sin (i * angle));

		bottom.Set(top.x,
						top.y + height,
						top.z);

		HC_Compute_Coordinates(".", "viewpoint", &bottom, "world", &bottom);
		HC_Compute_Coordinates(".", "viewpoint", &top, "world", &top);

		HC_Insert_Line(top.x, top.y, top.z, bottom.x, bottom.y, bottom.z);
	}
}



// insert a wireframe cone into the contruction plane using the viewpoint
// to world xform to determine the points
void Create_Wireframe_Cone(Point center, float radius, float height)
{
	Create_Axis_Circle(center, radius, XZ);

	float angle = 2.0f * PI / 8.0f;

	for (int i = 0; i < 8; i++) {
		Point top, bottom;
		top.Set(center.x + radius * (float)cos (i * angle),
						center.y,
						center.z + radius * (float)sin (i * angle));
		bottom.Set(center.x,
						center.y + height,
						center.z);

		HC_Compute_Coordinates(".", "viewpoint", &bottom, "world", &bottom);
		HC_Compute_Coordinates(".", "viewpoint", &top, "world", &top);

		HC_Insert_Line(top.x, top.y, top.z, bottom.x, bottom.y, bottom.z);
	}
}


int report_ifail(int ifail,int expect)
{
	if ( ifail != expect )
		return 1;
	else
		return 0;
}


// render the specified parasolid bodies
void parasolid_render (int number_of_bodies, PK_BODY_t *bodies)
{
	PK_ERROR_code_t				error_code;
	PK_TRANSF_t					*topol_transfs;
	PK_TRANSF_t					view_transf;
	PK_TOPOL_render_facet_o_t	rendering_options;
	PK_TOPOL_render_line_o_t	line_Options;
	PK_TOPOL_facet_mesh_o_t     generation_settings;
    PK_TOPOL_render_facet_go_o_t	go_options;

	topol_transfs = NULTAG;
	view_transf = NULTAG;
	
	if (number_of_bodies < 1) return;
	
	HP_Show_Render_Options(&line_Options, &go_options, &generation_settings);

	rendering_options.control = generation_settings;
	rendering_options.go_option = go_options;

	// render the facets (PK_Faces) If the compress_bodies flag is on, each Parasolid body
	// will be mapped to a HOOPS shell; otherwise, each Parasolid PK_Face will
	// be mapped to a HOOPS shell
	error_code = PK_TOPOL_render_facet(number_of_bodies, bodies, 
							topol_transfs, view_transf, &rendering_options);
	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed rendering cylinder faces\n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);
	}

	// render the lines (PK_Edges)  Each Parsolid PK_Edge will be mapped
	// to a HOOPS line, polyline, arc, ellipse, circular arc, or elliptical arc
	error_code = PK_TOPOL_render_line(number_of_bodies, bodies, 
							topol_transfs, view_transf, &line_Options);

	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed rendering cylinder edges \n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);
	}
}

// create a parasolid sphere and return the sphere's body id
PK_BODY_t parasolid_sphere(Point	& center, float	radius)
{
	PK_ERROR_code_t		error_code;

	// sphere creation parameters
	PK_AXIS2_sf_t	sphere_basis_set;
	PK_BODY_t		sphere_body;

	// rendering options
	int							number_of_entities;
	PK_TOPOL_t					entities[1];

	//  creating sphere
	sphere_basis_set.location.coord[0] = (double) center.x;
	sphere_basis_set.location.coord[1] = (double) center.y;
	sphere_basis_set.location.coord[2] = (double) center.z;

	sphere_basis_set.axis.coord[0] = 0.0;
	sphere_basis_set.axis.coord[1] = 1.0;
	sphere_basis_set.axis.coord[2] = 0.0;

	sphere_basis_set.ref_direction.coord[0] = 1.0;
	sphere_basis_set.ref_direction.coord[1] = 0.0;
	sphere_basis_set.ref_direction.coord[2] = 0.0;
	
	error_code = PK_BODY_create_solid_sphere(radius,  &sphere_basis_set, &sphere_body);

	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed rendering sphere \n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);
		return 0;
	}

	number_of_entities = 1;
	entities[0] = sphere_body;
	parasolid_render(number_of_entities, entities);


	return sphere_body;
}


// create a parasolid cylinder and return the cylinder's body id
PK_BODY_t parasolid_cylinder(Point	& center, Vector radius_vector, Vector height_vector)
{
	PK_ERROR_code_t		error_code;

	// sphere creation parameters
	PK_AXIS2_sf_t	cyl_basis_set;
	PK_BODY_t		cyl_body;

	// rendering options
	int				number_of_entities;
	PK_TOPOL_t		entities[1];
	float			radius, height;
	double	view[3], up[3];
	Point	add, subtract;

	Get_Viewplane_Vectors(view, up);

	radius = (float) HC_Compute_Vector_Length(&radius_vector);
	height = (float) HC_Compute_Vector_Length(&height_vector);

	// to determine orientation add and subtract the viewplane up vector 
	// and the cylinder up vector.  By measuring the size of each vector
	// we can determine the direction of the cylinder

	add.Set(height_vector.x + (float) up[0],
			height_vector.y + (float) up[1],
			height_vector.z + (float) up[2]);

	subtract.Set(height_vector.x - (float) up[0],
			height_vector.y - (float) up[1],
			height_vector.z - (float) up[2]);
	
	cyl_basis_set.location.coord[0] = (double) center.x;
	cyl_basis_set.location.coord[1] = (double) center.y;
	cyl_basis_set.location.coord[2] = (double) center.z;

	if (HC_Compute_Vector_Length(&add) > HC_Compute_Vector_Length(&subtract)){
		cyl_basis_set.axis.coord[0] = up[0];
		cyl_basis_set.axis.coord[1] = up[1];
		cyl_basis_set.axis.coord[2] = up[2];
	}
	else
	{
		cyl_basis_set.axis.coord[0] = -up[0];
		cyl_basis_set.axis.coord[1] = -up[1];
		cyl_basis_set.axis.coord[2] = -up[2];
	}

	cyl_basis_set.ref_direction.coord[0] = view[0];
	cyl_basis_set.ref_direction.coord[1] = view[1];
	cyl_basis_set.ref_direction.coord[2] = view[2];
		
	error_code = PK_BODY_create_solid_cyl(radius, height, &cyl_basis_set, &cyl_body);

	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed rendering cylinder\n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);
		return 0;
	}

	// rendering function
	number_of_entities = 1;
	entities[0] = cyl_body;
	parasolid_render(number_of_entities, entities);

	return cyl_body;
}


// create a parasolid cone and return the cone's body id
PK_BODY_t parasolid_cone(Point	& center, Vector radius_vector, Vector height_vector)
{
	PK_ERROR_code_t		error_code;
	double				angle;

	// cone creation parameters
	PK_AXIS2_sf_t	cone_basis_set;
	PK_BODY_t		cone_body;

	// rendering options
	int							number_of_entities;
	PK_TOPOL_t					entities[1];

	cone_basis_set.location.coord[0] = center.x;
	cone_basis_set.location.coord[1] = center.y;
	cone_basis_set.location.coord[2] = center.z;

	float		radius, height;
	double		view[3], up[3];
	Point		add, subtract;

	Get_Viewplane_Vectors(view, up);

	radius = (float) HC_Compute_Vector_Length(&radius_vector);
	height = (float) HC_Compute_Vector_Length(&height_vector);

	// to determine orientation add and subtract the viewplane up vector 
	// and the cylinder up vector.  By measuring the size of each vector
	// we can determine the direction of the cylinder

	add.Set(height_vector.x + (float) up[0],
			height_vector.y + (float) up[1],
			height_vector.z + (float) up[2]);

	subtract.Set(height_vector.x - (float) up[0],
				height_vector.y - (float) up[1],
				height_vector.z - (float) up[2]);
	
	cone_basis_set.location.coord[0] = (double) center.x + (double) height_vector.x;
	cone_basis_set.location.coord[1] = (double) center.y + (double) height_vector.y;
	cone_basis_set.location.coord[2] = (double) center.z + (double) height_vector.z;

	if (HC_Compute_Vector_Length(&add) < HC_Compute_Vector_Length(&subtract)){
		cone_basis_set.axis.coord[0] = up[0];
		cone_basis_set.axis.coord[1] = up[1];
		cone_basis_set.axis.coord[2] = up[2];
	}
	else
	{
		cone_basis_set.axis.coord[0] = -up[0];
		cone_basis_set.axis.coord[1] = -up[1];
		cone_basis_set.axis.coord[2] = -up[2];
	}

	cone_basis_set.ref_direction.coord[0] = view[0];
	cone_basis_set.ref_direction.coord[1] = view[1];
	cone_basis_set.ref_direction.coord[2] = view[2];
	
	angle = atan(radius/height);

	error_code = PK_BODY_create_solid_cone(0.0, height, angle, &cone_basis_set, &cone_body);

	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed rendering cone\n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);
		return 0;
	}

	// render bodies
	number_of_entities = 1;
	entities[0] = cone_body;
	parasolid_render(number_of_entities, entities);

	return cone_body;
}


// calls the appropriate boolean function based on the target and tools
// and rerenders the resulting bodies
void	parasolid_boolean(PK_boolean_function_t	operation, PK_BODY_t target, 
							  int number_of_tools, PK_BODY_t *tools)
{
	PK_ERROR_code_t		error_code;
	PK_BODY_boolean_o_t options;
	int                 number_of_bodies;
	PK_BODY_t          *bodies;

	assert(target);
	assert(number_of_tools > 0);

	PK_BODY_boolean_o_m(options);
    options.function = operation;

	// perform the boolean operation
	error_code = PK_BODY_boolean(target, number_of_tools, tools, &options, &number_of_bodies, &bodies);

	if ( report_ifail(error_code,0) != 0){
		char	error_message[256];
		sprintf(error_message, "failed during boolean operation \n  Error Code : %d", error_code);
		MessageBox(NULL, error_message, NULL, MB_OK);

	}

	// render bodies
	parasolid_render(number_of_bodies, bodies);
	if (number_of_bodies > 0)
		PK_MEMORY_free(bodies);

}
